/*
 * main.c - SubAloud subtitles reader
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <glib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "videosa.h"
#include "sonic.h"

char *module;
char *voice;
int phone_type;
char *movie_name;
char *test_mode;
char *sub_name;
int module_feat;
double min_rate=1.0;
double max_rate=1.0;
double use_fps;
int use_fps_set=0;

int modpipe_from[2],modpipe_to[2];
int module_pid;
FILE *from_module,*to_module;
char *audio_driver;
char *audio_param;
char *audio_device;

struct subtitle *sublist,**subarray;
int subarray_len;
int subarray_pos=-1;

static GThread *socket_Thread;

void helpme(char *fname,int ecode)
{
    fprintf(stderr,"Usage: %s <options> [-- <mplayer_options>]\n",fname);
    fprintf(stderr,"Options are:\n\
  -M, --module=name      - module name (required)\n\
  -V, --voice=name       - voice name\n\
  -m, --movie=path       - path to movie file\n\
  -r, --rate=n           - nominal speech rate (1.0 .. 2.0)\n\
  -R, --max-rate=n       - maximum speech rate (1.0 .. 2.0)\n\
  -T, --test=string      - test mode\n\
  -s, --subtitles=path   - subtitles file name\n\
  -l, --sublang=code     - 2-characters code of subtitles language\n\
                           for encoding autodetection (default locale)\n\
  -e, --subencoding=name - subtitles encoding (default autodetect)\n\
  -A, --volume=value     - initial volume (0 .. 100)\n\
  -D, --delay=value      - initial audio delay (in deciseconds,\n\
                           may be negative)\n\
  -S, --stdin            - do not start mplayer, read stdin instead\n\
  -P, --pass-thru        - copy mplayer output to stdout\n\
  -F, --fps=value        - frames per second (10.0 .. 80.0)\n\
  -a, --audio=driver     - audio driver in format driver:argument\n\
        Drivers are:\n\
         pulse:server         - pulseaudio internal driver\n\
         pulse:server:sink    - pulseaudio internal driver\n\
         pulse::sink          - pulseaudio internal driver\n\
         ao-pulse:server      - pulseaudio libao driver\n\
         ao-pulse:server:sink - pulseaudio libao driver\n\
         ao-pulse::sink       - pulseaudio libao driver\n\
         alsa:device          - alsa libao driver\n\
         oss:device           - OSS libao driver\n\
         nas:server           - NAS libao driver\n\
        Argument (with preceeding colon) may be omitted\n\
Module specific options (may be ignored by some modules):\n\
      --mod-pitch=value  - pitch (0.5 .. 2.0)\n\
      --mod-voice=path   - path to module voice file (f.ex. mbrola database)\n\
      --mod-data=path    - path to module data file (f.ex. txt2pho data)\n\
  -L, --mod-lang=string  - foreign language modifiers for module NLP\n\
      --mod-theme=string - themes for module NLP\n\
      --mod-dict=path    - pronunciation dictionary (module specific)\n\
      --mod-pre=type     - text preprocessor\n");

    exit(ecode);
}
int volume=100;

void simple_play_stretched(short *wave,int samples,int freq,double sonic_rate)
{
    short buf[1024];
    int err=0;
    int last=0;
    sonicStream sonic;
    sonic=sonicCreateStream(freq,1);
    sonicSetSpeed(sonic,sonic_rate);
    while (samples > 0 && !err) {
	int n=4096;
	if (n > samples) n=samples;
        sonicWriteShortToStream(sonic,wave,n);
        wave+=n;
        samples-=n;
	while (!err) {
            n=sonicReadShortFromStream(sonic,buf+last,1024-last);
            if (n<=0) break;
	    if (n+last < 1024) {
		last=n+last;
		break;
	    }
	    n += last;
	    last=0;
	    if (!audio_play_sfr(buf,n,freq)) {
		err=1;
		break;
	    }
	}
    }
    if (last > 0) {
	audio_play_sfr(buf,last,freq);
    }
    audio_play_sfr(NULL,0,freq);
    sonicDestroyStream(sonic);
}


void simple_play(short *wave,int samples,int freq,int use_sonic,double sonic_rate)
{
    if (use_sonic) {
	simple_play_stretched(wave,samples,freq,sonic_rate);
	return;
    }
    while (samples > 0) {
	int n;
	if (samples < 2048) n=samples;
	else n=1024;
	if (!audio_play_sfr(wave,n,freq)) break;
	wave += n;
	samples -= n;
    }
    audio_play_sfr(NULL,0,freq);
}


char *read_fline(FILE *f)
{
    static GString *gs=NULL;
    int znak;
    if (!gs) gs=g_string_sized_new(1024);
    g_string_assign(gs,"");
    for (;;) {
	znak=fgetc(f);
	if (znak == EOF) return NULL;
	g_string_append_c(gs,znak);
	if (znak=='\n') break;
    }
    return gs->str;
}

int read_module(char **outptr,int *outbytes,int *extra)
{
    GString *gs=NULL;
    char *line;
    int rc;
    
    *outptr=NULL;
    if (outbytes) *outbytes=0;
    if (extra) *extra=0;
    
    line=read_fline(from_module);
    if (!line) return 0;
    rc=strtol(line,&line,10);
    if (rc >= 300 && rc <=399) {
	char *c=strchr(line,':');
	int len,freq;
	if (!c) return 0;
	c++;
	freq=strtol(c,&c,10);
	if (*c++!=':') return 0;
	len=strtol(c,&c,10)*2;
	c=g_malloc(len);
	if (fread(c,1,len,from_module)!=len) {
	    g_free(c);
	    return 0;
	}
	if (fgetc(from_module) != '\n') {
	    g_free(c);
	    return 0;
	}
	line=read_fline(from_module);
	if (!line || strtol(line,NULL,10)!=rc) {
	    g_free(c);
	    return 0;
	}
	*outptr=c;
	if (outbytes) *outbytes=len;
	if (extra) *extra=freq;
	return rc;
    }
    for (;;) {
	int cont=(*line++ == '-');
	if (!gs) gs=g_string_new(line);
	else g_string_append(gs,line);
	if (!cont) break;
	line=read_fline(from_module);
	if (!line || strtol(line,&line,10)!=rc) {
	    g_string_free(gs,TRUE);
	    return 0;
	}
    }
    if (outbytes) *outbytes=gs->len;
    *outptr=g_string_free(gs,0);
    return rc;
}

void send_command_body(char *str)
{
    char *c;
    for (;*str;) {
	c=strstr(str,"\n.");
	if (!c) {
	    int n=strlen(str);
	    fwrite(str,1,n,to_module);
	    if (str[n-1] != '\n') fputc('\n',to_module);
	    break;
	}
	c++;
	fwrite(str,1,c-str,to_module);
	fputc('.',to_module);
	str=c;
    }
    fputs(".\n",to_module);
    fflush(to_module);
}

static char *mod_data,*mod_voice,*mod_lang,*mod_theme,
    *mod_dict,*mod_preproc;

static int mod_pitch_set;
static double mod_pitch;

void start_module(void)
{
    char module_path[256],bf1[16];
    char *args[32];
    int nargs=0;
    int rc;
    char *iline,*cs,*ds,*es;
    
    sprintf(module_path,MODPATH"/vox_%s",module);
    if (!g_file_test(module_path,G_FILE_TEST_IS_EXECUTABLE)) {
	fprintf(stderr,"Cannot find module %s\n",module);
	exit(1);
    }
    if (pipe(modpipe_from) || pipe(modpipe_to)) {
	perror("pipe");
	exit(1);
    }
    module_pid=fork();
    if (module_pid<0) {
	perror("fork");
	exit(1);
    }
    if (!module_pid) {
	args[nargs++]=module;
	if (mod_pitch_set) {
	    sprintf(bf1,"%.2f",mod_pitch);
	    args[nargs++]="--mbrola-pitch";
	    args[nargs++]=bf1;
	}
	if (mod_voice) {
	    args[nargs++]="--mbrola-voice";
	    args[nargs++]=mod_voice;
	}
	if (mod_data) {
	    args[nargs++]="--datapath";
	    args[nargs++]=mod_data;
	}
	if (mod_lang) {
	    args[nargs++]="--languages";
	    args[nargs++]=mod_lang;
	}
	if (mod_theme) {
	    args[nargs++]="--themes";
	    args[nargs++]=mod_theme;
	}
	if (mod_dict) {
	    args[nargs++]="--dict";
	    args[nargs++]=mod_dict;
	}
	if (mod_preproc) {
	    args[nargs++]="--pre";
	    args[nargs++]=mod_preproc;
	}
	    
	if (voice) args[nargs++]=voice;
	args[nargs++]=0;
	close(modpipe_from[0]);
	close(modpipe_to[1]);
	dup2(modpipe_from[1],1);
	dup2(modpipe_to[0],0);
	execv(module_path,args);
	exit(1);
    }
    close(modpipe_from[1]);
    close(modpipe_to[0]);
    from_module=fdopen(modpipe_from[0],"r");
    to_module=fdopen(modpipe_to[1],"w");
    rc=read_module(&iline,NULL,NULL);
    if (rc != 200) {
	fprintf(stderr,"ERROR %s\n",iline);
	exit(1);
    }
    for (cs=es=iline;cs && *cs;cs=es) {
	es=strchr(cs,'\n');
	if (es) *es++=0;
	ds=strchr(cs,'=');
	if (ds) *ds++=0;
	if (!strcmp(cs,"wave")) {
	    module_feat|=VSA_MFEAT_WAVE;
	}
	else if (!strcmp(cs,"rwave")) {
	    module_feat|=VSA_MFEAT_RWAVE;
	}
	else if (!strcmp(cs,"join")) {
	    module_feat|=VSA_MFEAT_JOIN;
	}
	else if (!strcmp(cs,"trans")) {
	    module_feat|=VSA_MFEAT_TRANS;
	    if (!ds) phone_type=VSA_PHMODE_PRIVATE;
	    else if (!strcmp(ds,"mbrola")) phone_type=VSA_PHMODE_MBROLA;
	    else if (!strcmp(ds,"orto")) phone_type=VSA_PHMODE_ORTO;
	    else if (!strcmp(ds,"sampa")) phone_type=VSA_PHMODE_SAMPA;
	    else if (!strcmp(ds,"ipa")) phone_type=VSA_PHMODE_IPA;
	    else phone_type=VSA_PHMODE_PRIVATE;
	}
    }
    g_free(iline);
}

char *get_translation(char *str)
{
    int rc;char *iline;
    if (!(module_feat & VSA_MFEAT_TRANS)) return NULL;
    fprintf(to_module,"trans %s\n",str);
    fflush(to_module);
    rc=read_module(&iline,NULL,NULL);
    if (rc < 200 || rc > 299) {
	g_free(iline);
	return NULL;
    }
    return iline;
}

short *get_wave(char *str,int *samples,int *freq)
{
    short *wave;
    int len,rc;
    if (!(module_feat & VSA_MFEAT_WAVE)) return NULL;
    fprintf(to_module,"@wave\n");
    send_command_body(str);
    rc=read_module((char **)&wave,&len,freq);
    if (rc < 300 || rc > 399) {
	printf("ERROR %s\n",(char *)wave);
	g_free(wave);
	return NULL;
    }
    *samples=len/2;
    return wave;
}

short *get_rwave(char *str,int *samples,int *freq,double rate)
{
    short *wave;
    int len,rc;
    if (!(module_feat & VSA_MFEAT_RWAVE)) return NULL;
    fprintf(to_module,"@rwave %f\n",rate);
    send_command_body(str);
    rc=read_module((char **)&wave,&len,freq);
    if (rc < 300 || rc > 399) {
	printf("ERROR %s\n",(char *)wave);
	g_free(wave);
	return NULL;
    }
    *samples=len/2;
    return wave;
}


void test_me(void)
{
    char *s;int len,freq;short *wave;
    if (module_feat & VSA_MFEAT_TRANS) {
	s=get_translation(test_mode);
    }
    else {
	s=test_mode;
    }
    if (!s) return;
    if (module_feat & VSA_MFEAT_RWAVE) {
	wave=get_rwave(s,&len,&freq,min_rate);
	if (wave) {
	    printf("Wave len %d freq %d\n",len,freq);
	    simple_play(wave,len,freq,0,0.0);
	    g_free(wave);
	}
	wave=get_rwave(s,&len,&freq,max_rate);
	if (wave) {
	    printf("Wave len %d freq %d\n",len,freq);
	    volumize(wave,len);
	    simple_play(wave,len,freq,0,0.0);
	    g_free(wave);
	}
	exit(0);
    }
    wave=get_wave(s,&len,&freq);
    if (wave) {
	printf("Wave len %d freq %d\n",len,freq);
	volumize(wave,len);
	simple_play_stretched(wave,len,freq,min_rate);
	simple_play_stretched(wave,len,freq,max_rate);
    }
}

void volumize(short *wave,int len)
{
    int i;
    if (volume <100 && volume > 0) {
	for (i=0;i<len;i++) wave[i]=(wave[i] * volume)/100;
    }
}

int socket_fd;
struct sockaddr_un my_socket_addr;
int synchro_sub;
int sub_delay;

gpointer socket_ThreadFunc(gpointer data)
{
    char buf[256];
    char *abstract_name="-SubAloud.ctrl";
    socket_fd=socket(AF_UNIX, SOCK_STREAM, 0);
    memset(&my_socket_addr, 0, sizeof(struct sockaddr_un));
    my_socket_addr.sun_family = AF_UNIX;
    strcpy(my_socket_addr.sun_path, abstract_name);
    int len_unix=SUN_LEN(&my_socket_addr);
    my_socket_addr.sun_path[0]=0;
    if (bind(socket_fd, (struct sockaddr *) &my_socket_addr,
                   len_unix) == -1) {
	        perror("SubAloud can't bind to socket");
		return NULL;
	    }
    if (listen(socket_fd, 50) == -1) {
	perror("SubAloud Listen");
	return NULL;
    }
	       
    for (;;) {
	int fd=accept(socket_fd,NULL,NULL);
	if (fd < 0) break;
	FILE *f=fdopen(fd,"w+");
	if (fgets(buf,256,f)) {
	    fprintf(stderr,"Socket received [%s]\n",buf);
	    if (buf[0] == 'd') {
		if (buf[1] == '+' && (!buf[2] ||  !isdigit(buf[2]))) {
		    sub_delay=sub_delay+1;
		}
		else if (buf[1] == '-' && (!buf[2] ||  !isdigit(buf[2]))) {
		    sub_delay=sub_delay-1;
		}
		else {
		    sub_delay=strtol(buf+1,NULL,10);
		}
		fprintf(f,"+d=%d\n",sub_delay);
	    }
	    else if (buf[0] == 'g') {
		fprintf(f,"d=%d v=%d\n",sub_delay,volume);
	    }
	    else if (buf[0] == 'v') {
		int v;
		if (buf[1] == '+') {
		    v=volume+5;
		    if (v>100) v=100;
		}
		else if (buf[1] == '-') {
		    v=volume-5;
		    if (v<0) v=0;
		}
		else {
		    v=strtol(buf+1,NULL,10);
		}
		if (v<=0 && v >= 100) {
		    fprintf(f,"-illegal value\n");
		}
		else {
		    volume=v;
		    fprintf(f,"+v=%d\n",volume);
		}
	    }
	    else if (buf[0] == 's') {
		synchro_sub=1;
		fprintf(f,"+ok\n");
	    }
	    else if (buf[0] == 'p') {
		fprintf(f,"+p=%d\n",current_decisec);
	    }
	    else {
		fprintf(f,"-bad command\n");
	    }
	}
	fclose(f);
    }
    close(socket_fd);
    return NULL;
}

void start_socket_thread(void)
{
#if !GLIB_CHECK_VERSION(2,32,0)
    socket_Thread=g_thread_create(socket_ThreadFunc,NULL,TRUE,NULL);
#else
    socket_Thread=g_thread_new("vms_socket",socket_ThreadFunc,NULL);
#endif

}

int main(int argc,char *argv[])
{
    char *optstr="M:V:hm:T:r:R:s:A:D:SPF:a:";
    char *sublang=NULL,*subenc=NULL;
    int isframes=0;
    struct option options[]={
	{"module",1,NULL,'M'},
	{"voice",1,NULL,'V'},
	{"movie",1,NULL,'m'},
	{"test",1,NULL,'T'},
	{"help",0,NULL,'h'},
	{"rate",0,NULL,'r'},
	{"max-rate",0,NULL,'R'},
	{"subtitles",1,NULL,'s'},
	{"sublang",1,NULL,'l'},
	{"subencoding",1,NULL,'e'},
	{"volume",1,NULL,'A'},
	{"delay",1,NULL,'D'},
	{"stdin",0,NULL,'S'},
	{"pass-thru",0,NULL,'P'},
	{"fps",1,NULL,'F'},
	{"audio",1,NULL,'a'},
	{"mod-pitch",1,NULL,0x10},
	{"mod-voice",1,NULL,0x11},
	{"mod-data",1,NULL,0x12},
	{"mod-lang",1,NULL,'L'},
	{"mod-theme",1,NULL,0x13},
	{"mod-dict",1,NULL,0x14},
	{"mod-pre",1,NULL,0x15},
	{NULL,0,NULL,0}};
    
    fprintf(stderr,"SubAloud v. " VERSION " (C) 2013 Bohdan R. Rau\n");
    for (;;) {
	int opt=getopt_long(argc,argv,optstr,options,NULL);
	double dr;
	if (opt < 0) break;
	switch(opt) {
	    
	    case 0x10:
	    dr=strtod(optarg,NULL);
	    if ((dr < 0.5) || (dr > 2.0)) helpme(argv[0],1);
	    mod_pitch=dr;
	    mod_pitch_set=1;
	    break;
	    
	    case 0x11:
	    mod_voice=strdup(optarg);
	    continue;
	    
	    case 0x12:
	    mod_data=strdup(optarg);
	    continue;
	    
	    case 0x13:
	    mod_theme=strdup(optarg);
	    continue;
	    
	    case 'L':
	    mod_lang=strdup(optarg);
	    continue;
	    
	    case 0x14:
	    mod_dict=strdup(optarg);
	    continue;

	    case 0x15:
	    mod_preproc=strdup(optarg);
	    continue;
	    
	    case 'A':
	    volume=strtol(optarg,NULL,10);
	    if (volume > 100 || volume < 0) volume=100;
	    break;
	    
	    case 'M':
	    module=strdup(optarg);
	    break;
	    
	    case 'V':
	    voice=strdup(optarg);
	    break;
	    
	    case 'h':
	    helpme(argv[0],0);
	    
	    case 'T':
	    test_mode=strdup(optarg);
	    break;
	    
	    case 'r':
	    case 'R':
	    dr=strtod(optarg,NULL);
	    if ((dr < 1.0) || (dr > 2.0)) helpme(argv[0],1);
	    if (opt == 'r') min_rate=dr;
	    else max_rate=dr;
	    
	    case 'm':
	    movie_name=strdup(optarg);
	    break;
	    
	    case 's':
	    sub_name=strdup(optarg);
	    break;
	    
	    case 'l':
	    sublang=strdup(optarg);
	    break;

	    case 'e':
	    subenc=strdup(optarg);
	    break;

	    case 'a':
	    audio_driver=strdup(optarg);
	    break;

	    case 'D':
	    sub_delay=strtol(optarg,NULL,10);
	    break;
	    
	    case 'S':
	    mplayer_stdin = 1;
	    break;
	    
	    case 'P':
	    mplayer_pass = 1;
	    break;
	    
	    case 'F':
	    dr=strtod(optarg,NULL);
	    if (dr < 10.0 || dr > 80.0) helpme(argv[0],1);
	    use_fps=dr;
	    use_fps_set=1;
	    break;

	    default:
	    helpme(argv[0],1);
	}
    }
    fprintf(stderr,"Arguments parsed\n");
    use_pulse=1;
    if (audio_driver) {
	char *c=strchr(audio_driver,':');
	if (c) *c++=0;
	audio_param=c;
	if (strcmp(audio_driver,"pulse")) use_pulse=0;
	if (!strcmp(audio_driver,"pulse") || !strcmp(audio_driver,"ao-pulse")) {
	    c=strchr(audio_param,':');
	    if (c) {
		*c++=0;
		audio_device=c;
		if (!*audio_param) audio_param="default";
	    }
	}
    }
    if (max_rate < min_rate) max_rate=min_rate;
    if (!module) helpme(argv[0],1);
#if !GLIB_CHECK_VERSION(2,32,0)
    fprintf(stderr,"Initialising threads\n");
    g_thread_init(NULL);
    
#endif
    
    if (!test_mode) {
	fprintf(stderr,"Starting socket thread\n");
	start_socket_thread();
    }
    fprintf(stderr,"Starting module %s\n",module);
    start_module();
    fprintf(stderr,"Module started\n");
    if (test_mode) {
	test_me();
	//quit_module();
	exit(0);
    }
    if (!movie_name && !sub_name) {
	helpme(argv[0],1);
    }
    fprintf(stderr,"Reading subtitles\n");
    sublist=subtitle_read_buffer(sub_name,movie_name,subenc,sublang,&isframes);
    if (!sublist) {
	fprintf(stderr,"Cannot read subtitles\n");
	exit(1);
    }
    fprintf(stderr,"Subtitles OK\n");
    if (!isframes) {
	preproc_subtitles();
	make_subarray();
    }
    start_mplayer(argc-optind,argv+optind,isframes);
    if (isframes) {
	int isf;
	struct subtitle *s;
	for (isf=0;isf<500 && !fps_set;isf++) {
	    usleep(10000);
	}
	if (!fps_set) {
	    fprintf(stderr,"FPS not set\n");
	    exit(1);
	}
	for (s=sublist;s;s=s->next) {
	    s->decisecs=(10.0 * s->frames)/fps;
	}
	preproc_subtitles();
	make_subarray();
    }
    fprintf(stderr,"Subtitles %d\n",subarray_len);
    subarray_pos=-1;
    start_fetcher();
    play_subtitles();
    return 0;
}
